package TreeHasher;

# Dirty module containing all read-only code shared between the hashsum computer# and the extractor. The extractor shall not possibly modify anything. Hence,
# all dangerous code is in the separate hash writer program.

use File::Basename qw(basename dirname fileparse);
use File::Spec::Functions qw(abs2rel rel2abs splitdir catdir);
use Cwd qw(cwd);
use Exporter;

use strict;

our @ISA = 'Exporter';
our @EXPORT = qw(
	$sizelimit $ckname $hash_alg $reclimit
	$base $ckdir $ckbase $wrk $timestamp
	$verbose
	match_hashfile hashfile_name
	is_hashfile
	is_small_file is_big_file
	last_dirhash last_hashtime
	files dirs dir_is_empty
	open_read get_hashlist
	parse_hashline hashline
	dir_is_fresher write_hashes dir_is_static
	msg
);

# Directory under base where the checksum tree lives.
our $ckname = '.checksum';
# Only SHA-based sums are supported (regexes use sha\d+).
our $hash_alg = 'sha512';
# Limit directory recursions. There are always bugs.
our $reclimit = 100;
# Verbosity levels: 0=quiet, 1=normal, 2=chatty.
our $verbose = 0;

my $hash_prog;

our $base;
our $ckdir;
our $ckbase;
our $wrk;
our $timestamp;
our %cache;
# Cache that many lines of hash file data.
# This is the lazy route instead of actually computing memory
# footprint. With 256 bytes per line, we're at 24 MiB. We should have
# 100 MiB of RAM for such a core task ...
# This limit means that things start getting dropped if it is reached
# before adding a new file. So there will be somewhat more lines, usually.
# TODO: Count memory properly using the storage size of the cache.
our $cachelines = 100000;
# History of cached files in order, along with maching line count.
our @cache_files;
our @cache_lines;
# Regexes for static directories.
our @staticex;

sub init
{
	# Base directory we're dealing with (p.ex. /home).
	$base = rel2abs(shift);
	my $makeckdir = shift;
	# Get the path of the hashing program once, otherwise there is
	# overhead looking for it in $PATH all the time.
	# Of  course, one could use some more Perl-ish checksum computation,
	# but I'd like to keep the requirements on Perl modules minimal.
	$hash_prog = `which ${hash_alg}sum`;
	chomp($hash_prog);
	$hash_prog ne ""
		or die "$0: cannot find the checksum binary for $hash_alg\n";
	# Checksum directory has a special place in $base/.checksum.
	# The actual base directory is subdir 'base' inside that, to be able to
	# have a meaningful name for the  toplevel checksum file.
	# You translate from checksum file to real-life directory by stripping the hash
	# suffixes and s:^$ckbase:$base:;
	$ckdir  = $base.'/'.$ckname;
	# Create the toplevel checksum directory before creating our timestamp,
	# so that there is not a bogus re-hashing of the toplevel dir triggered
	# on the next run.
	if(not -e $ckdir and $makeckdir)
	{
		print "Creating $ckdir and waiting a bit to ensure timestamp monotony.\n"
			if $verbose;
		mkdir($ckdir)
			or die "$0: cannot create $ckdir: $!\n";
		sleep(2); # Really ensure timestamp is increased.
	}
	$ckbase = $ckdir.'/base';
	@staticex = ();
	if(open(my $fh, '<', "$ckdir/staticdirs"))
	{
		while(<$fh>)
		{
			next
				if(m/^\s*#/ or not m/\S/);
			chomp;
			push(@staticex, qr:$_:);
		}
	}
	$wrk = cwd();
	$timestamp = time();
}

# Support some alignment to better match up paths following that
# prefix.
sub msg
{
	my $prefix = shift;
	return sprintf("%-15s", $prefix.':').join('',@_)."\n";
}

sub match_hashfile
{
	my $file = shift;
	my $match = $file =~ /^(.+)\.(sha\d+)(\.(\d+)|)$/;
	# Fun way to return optional stuff: via aliased input array.
	# Actual return value always is a scalar for successful match.
	# The @_ size checks are not really needed, but it looks cleaner to me.
	if($match and $#_ >= 0)
	{
		$_[0] = $1; # directory name
		if($#_ >= 1)
		{
			$_[1] = $2; # hash suffix
			if($#_ >= 2)
			{
				$_[2] = $4; # timestamp
			}
		}
	}
	return $match;
}


sub hashfile_name
{
	my $name = shift;
	$name =~ s:/\.$::; # blabla/. = blabla
	$name .= '.'.$_[0]; # Hash algo needs to be present!
	# The rest is optional.
	$name .= '.'.$_[1]  # timestamp
		if(defined $_[1] and $_[1] ne '');
	return $name;
}

sub is_hashfile
{
	return (-f $_[0] and match_hashfile($_[0]));
}

# Find the matching directory hash file given the relative directory
# path.
# Optionally restrict time range (set both to same value for selcting
# specific time stamp).
# Optionally restrict algorithm.
sub last_dirhash
{
	my $relname  = shift; # path relative to base
	my $mintime  = shift; # inclusive time range
	my $maxtime  = shift;
	my $alg      = shift; # hash algorithm
	my $dirname = $ckbase.'/'.$relname;
	$dirname =~ s:/\.$::;
	return undef # Not over the top.
		if abs2rel($dirname, $ckbase) =~ m:^\.\./:;
	my $basename = basename($dirname);
	   $dirname  =  dirname($dirname);
	my %cand;
	for(@{files($dirname)})
	{
		if(match_hashfile($_, my ($name, $hash, $stamp)))
		{
			next
				if($name ne $basename);
			next
				if(defined $alg and $hash ne $alg);
			next
				if(defined $mintime and $stamp < $mintime);
			next
				if(defined $maxtime and $stamp > $maxtime);
			# Now either collect all candidates with their time stamps as key
			# or just the one if a time was given.
			$cand{$stamp} = 
			{
				dirpath=>$dirname.'/'.$name
			,	hashpath=>$dirname.'/'.$_
			,	timestamp=>$stamp
			,	hashalg=>$hash
			};
		}
	}
	if(%cand)
	{
		# The freshest collected candidate is it.
		my $key;
		for(keys %cand){ $key = $_ if $_ > $key; }
		return $cand{$key};
	}
	else
	{
		return undef
	}
}

sub last_hashtime
{
	my $relname = shift;
	my $alg     = shift;
	my $lasthash = last_dirhash($relname, undef, undef, $alg);
	return $lasthash->{timestamp};
}

# Return true if something in the directory is fresher as the given time.
# Pessimistic view: If the mtime is equal the reference, we consider
# the directory also to be fresher. There is a whole second in which things
# can happen.

sub mtimecheck
{
	my $path = shift;
	my $mtime = (stat($path))[9];
	my $reftime = shift;
	if($mtime > $timestamp and defined $reftime)
	{
		# There are files with modification times that far in the future that
		# the hashes are always considered out of date. Also, what is touched
		# right now, since program start, is too hot to consider archived.
		# Actually, one could configure a certain minimum age ...
		print STDERR msg('too modern', $path);
		return 0;
	}
	# If reftime is undefined, anything is newer.
	return $mtime >= $reftime;
}

sub dir_is_static
{
	my $reldir = shift;
	for(@staticex)
	{
		return 1
			if $reldir =~ $_;
	}
	return 0;
}

sub dir_is_fresher
{
	my $dir = shift;
	my $reftime = shift;
	return 1
		unless defined $reftime;
	# Fresh if entries have been added/removed/renamed.
	return 1
		if mtimecheck($dir, $reftime);
	opendir(my $dh, $dir)
		or return 0;
	# Fresh if some contained plain file changed.
	while(my $e = readdir($dh))
	{
		next
			if($e eq '.' or $e eq '..');
		my $file = $dir.'/'.$e;
		next
			unless(not -l $file and -f $file);
		return 1
			if mtimecheck($file, $reftime);
	}
	return 0;
}

# Get list of files and direct sub-directories. This excludes symlinks
# because I do not want to compute checksums multiple times and also
# do not like the overall confusion that can be caused.

sub files
{
	my $parent = defined $_[0] ? $_[0] : '.';
	opendir(my $dh, $parent)
		or return [];
	my @files;
	while(my $f = readdir($dh))
	{
		push(@files, $f)
			if( -f $parent.'/'.$f and not -l $parent.'/'.$f);
	}
	return \@files;
}

sub dirs
{
	my $parent = defined $_[0] ? $_[0] : '.';
	opendir(my $dh, $parent)
		or return [];
	my @dirs;
	while(my $d = readdir($dh))
	{
		if( $d ne '.' and $d ne '..'
		    and -d $parent.'/'.$d and not -l $parent.'/'.$d )
		{
			push(@dirs, $d)
		}
	}
	return \@dirs;
}

sub dir_is_empty
{
	opendir(my $dh, defined $_[0] ? $_[0] : '.')
		or return 0;
	while(my $e = readdir($dh))
	{
		return 0
			unless($e eq '.' or $e eq '..');
	}
	return 1;
}

sub open_read
{
	my $file = shift;
	my $fh;
	open($fh, '<', $file)
		or die "$0: cannot open $file: $!\n";
	return $fh;
}

sub open_read_try
{
	my $file = shift;
	return open_read($file)
		if -e $_;
	return undef;
}

sub parse_hashline
{
	my ($hash, $hfile) = split('  ', $_[0], 2);
	chomp($hfile);
	my %entry = (hash=>$hash);
	($entry{file}, $entry{dir}) = fileparse($hfile);
	return \%entry;
}

sub hashline
{
	my $spec = shift;
	return $spec->{hash}
	.	'  '.(defined $_[0] ? $_[0] : $spec->{dir})
	.	(defined $_[1] ? $_[1] : $_->{file})
	.	"\n";
}

sub cache_shrink
{
	my $cached_lines = 0;
	for(@cache_lines){ $cached_lines += $_; }
	while($cached_lines >= $cachelines)
	{
		my $deadite = shift(@cache_files);
		my $deadlin = shift(@cache_lines);
		delete($cache{$deadite});
		$cached_lines -= $deadlin;
	}
}

sub cache_push
{
	my $path = shift;
	my $lines = shift;
	$cache{$path} = $lines;
	push(@cache_files, $path);
	push(@cache_lines, scalar @{$lines});
}

sub add_to_cache
{
	my $hashpath = shift;
	# Clean up old stuff.
	cache_shrink();
	# Load new stuff.
	my @lines;
	my $fh = open_read($hashpath);
	die "$0: cannot read: $hashpath\n"
		unless defined $fh;
	while(<$fh>)
	{
		push(@lines, parse_hashline($_));
	}
	cache_push($hashpath, \@lines);
	return \@lines;
}

sub get_hashlist
{
	my $hashpath = shift; # Full path to hash file.
	return exists $cache{$hashpath}
	?	$cache{$hashpath}
	:	add_to_cache($hashpath);
}

# Consider feeding multiple files at a time for some reduced program
# startup time. But the list of files might be too long and I don't
# really think that startup of sha512sum matters.
# Also, if individual files get pulled by the user, we might get
# confused if not dealing with them individually.
# Note: It is assumed that hashfile exists after calling this, even if it is
# empty!
sub write_hashes
{
	my $hashfile = shift;
	my $caching = shift;
	open(my $fh, '>', $hashfile)
		or die "$0: cannot write: $hashfile\n";
	for my $f (@_)
	{
		# Get exactly one hash
		unless(open(my $ch, '-|', $hash_prog, '--', $f))
		{
			print "skipping bad file: $f\n"
				if $verbose;
		}
		else
		{
			my @lines;
			while(my $line = <$ch>)
			{
				print {$fh} $line;
				push(@lines, parse_hashline($line))
					if $caching;
			}
			cache_push($hashfile, \@lines)
				if $caching;
		}
	}
}

1;
